#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

extern int h_errno;
extern int errno;
int p_bytes = 512;
int p_port = 21;
char *p_pass = NULL;
char *p_login = NULL;
struct timeval p_wait = { 3, 0 };
int p_at = 2;
int p_end = 1000;		/* large enough */
int p_cwait = 0;
int p_nologin = 0;
enum states { START, PREAUTH, LOGGED } p_state;

#define BUFMAX ((1<<16) + 10)
#define RBUF ((1 << 17) + 10)
#define DIE(X) die (X, __FILE__, __LINE__);

char *cmds[] = {
  "USER %s", /* USER <SP> <username> <CRLF> */
  "PASS %s", /* PASS <SP> <password> <CRLF> */
  "ACCT %s", /* ACCT <SP> <account-information> <CRLF> */
  "CWD %s", /* CWD  <SP> <pathname> <CRLF> */
  "CDUP", /* CDUP <CRLF> */
  "SMNT %s", /* SMNT <SP> <pathname> <CRLF> */
  "QUIT", /* QUIT <CRLF> */
  "REIN", /* REIN <CRLF> */
  "PORT %s", /* PORT <SP> <host-port> <CRLF> */
  "PASV", /* PASV <CRLF> */
  "TYPE %s", /* TYPE <SP> <type-code> <CRLF> */
  "STRU %s", /* STRU <SP> <structure-code> <CRLF> */
  "MODE %s", /* MODE <SP> <mode-code> <CRLF> */
  "RETR %s", /* RETR <SP> <pathname> <CRLF> */
  "STOR %s", /* STOR <SP> <pathname> <CRLF> */
  "STOU", /* STOU <CRLF> */
  "APPE %s", /* APPE <SP> <pathname> <CRLF> */
  "ALLO %s", /* ALLO <decimal-integer> [<SP> R <SP> <decimal-integer>] <CRLF> */
  "REST %s", /* REST <SP> <marker> <CRLF> */
  "RNFR %s", /* RNFR <SP> <pathname> <CRLF> */
  "RNTO %s", /* RNTO <SP> <pathname> <CRLF> */
  "ABOR", /* ABOR <CRLF> */
  "DELE %s", /* DELE <SP> <pathname> <CRLF> */
  "RMD %s", /* RMD  <SP> <pathname> <CRLF> */
  "MKD %s", /* MKD  <SP> <pathname> <CRLF> */
  "PWD %s", /* PWD  <CRLF> */
  "LIST %s", /* LIST [<SP> <pathname>] <CRLF> */
  "NLST %s", /* NLST [<SP> <pathname>] <CRLF> */
  "SITE %s", /* SITE <SP> <string> <CRLF> */
  "SYST", /* SYST <CRLF> */
  "STAT %s", /* STAT [<SP> <pathname>] <CRLF> */
  "HELP %s", /* HELP [<SP> <string>] <CRLF> */
  "NOOP", /* NOOP <CRLF> */
  NULL
};
/* external */
char *do_rle (char *s, size_t len);
/* here */
int child_donep (pid_t *pid);
char *make_big (char *buf, char *cmd, size_t bytes);
int printIfLine (char *rbuf, char **tbuf);
void parse_it (int sock, char *rbuf, char **tbuf, char *wbuf);

void die (char *err, char *f, int line) {
  int oe = errno;
  fprintf (stderr, "bad: ");
  errno = oe;
  perror (err);
  fprintf (stderr, "%s:%d\n", f, line);
  exit ((line % 30) + 1);
}
pid_t send_it (int socket, char *buf, size_t bytes) {
  static pid_t pid = 0;
#define OLINE 50
  char t[50];

  while (pid && !child_donep (&pid)) {
    printf ("Child %d not terminated, sleeping\n", (int)pid);
    sleep (1);
  }
  strncpy(t, buf, (bytes >= OLINE - 1)?OLINE -1:bytes);
  *(t + ((bytes >= OLINE - 1)?OLINE -1:bytes) - 1) = '\0';
  printf ("Try: %s\n", t);
  fflush(stdout);
  if ((pid = fork()) == 0) {	/* child */
    if (write(socket, buf, bytes) != bytes)
      DIE("write");
    exit(0);
  } else if (pid < 0) {		/* error */
    DIE ("fork");
  }
  return pid;
}

int makeconnect (char *host, int port) {
  struct	sockaddr_in hisctladdr, myaddr;
  struct hostent *hp = NULL;
  char hostnamebuf[RBUF];
  int sock;
  size_t alen = sizeof(myaddr);

  memset (&hisctladdr, 0, sizeof (hisctladdr));
  if ((hisctladdr.sin_addr.s_addr = inet_addr(host)) != -1) {
    hisctladdr.sin_family = AF_INET;
    (void) strncpy(hostnamebuf, host, sizeof(hostnamebuf));
  } else {
    hp = gethostbyname(host);
    if (hp == NULL) {
      fprintf(stderr, "bad: %s: (%d)\n", host, h_errno);
      exit (2);
    }
    hisctladdr.sin_family = hp->h_addrtype;
    memcpy((caddr_t)&hisctladdr.sin_addr, 
	   hp->h_addr_list[0], hp->h_length);
    (void) strncpy(hostnamebuf, hp->h_name, sizeof(hostnamebuf));
  }
  if ((sock = socket (hisctladdr.sin_family, SOCK_STREAM, 0)) < 0) {
    DIE("socket");
  }
  hisctladdr.sin_port = htons(port) ;
  
  if (connect(sock, (struct sockaddr *)&hisctladdr, sizeof (hisctladdr)) < 0) {
    DIE ("connect");
  }
  if (getsockname (sock, (struct sockaddr *)&myaddr, &alen) == 0) {
    printf ("Connected: %s.%d <=> %s.%d\n", inet_ntoa(myaddr.sin_addr), 
	    ntohs(myaddr.sin_port), host, port);
  } else
    DIE ("getsockname");
  return sock;
}

int child_donep (pid_t *pid) {
  int stat_loc;
  if (*pid == 0)
    return 1;
  if (waitpid(*pid, &stat_loc, WNOHANG) == *pid) {
    if (WIFEXITED(stat_loc) == 0 &&
	WIFSIGNALED(stat_loc) == 0)
      return 0;
  }
  *pid = 0;
  return 1;
}

void try_break (int sock, char *wbuf) {
  static int tries = 0;
  if (cmds[p_at] == NULL || p_at > p_end) {
    char *txt = "QUIT\r\n"; /* all through */

    if (++tries > 5)
      DIE("Tried too many times to QUIT");
    send_it (sock, txt, strlen(txt));
  } else {
    char *s = strchr (cmds[p_at], ' ');
    char *p;
    
    if (s == NULL) {	/* no arg */
      p = malloc(strlen(cmds[p_at]) + 2);
      if (p == NULL) {
	DIE("malloc");
      }
      strcpy (p, cmds[p_at]);	/* calculated */
      strcat (p, " ");	/* calculated */
    } else {		/* has arg */
      if ((p = strdup(cmds[p_at])) == NULL) {
	DIE ("strdup");
      }
      *(strchr(p, ' ')+1) = '\000'; /* does not fail */
    }
    send_it (sock, make_big(wbuf, p, p_bytes), p_bytes);
    free(p);
    p_at++;		/* increment here */
  }
}

void do_loop (int sock, char *rbuf, char **tbuf, char *wbuf) {
 fd_set rset;
 int got = 0;

 FD_ZERO(&rset);
 FD_SET(sock, &rset);

 while (1) {
   switch (select (sock + 1, &rset, NULL, NULL, &p_wait)) {
   case 0:			/* timeout */
     parse_it (sock, rbuf, tbuf, wbuf);
     try_break(sock, wbuf);
     break;
   case 1:			/* something to read */
     got = read (sock, *tbuf, RBUF - (*tbuf - rbuf));
     if (got == 0)
       return;			/* is this ok? EOF? */
     else if (got < 0)
       return;
     *tbuf += got;
     parse_it (sock, rbuf, tbuf, wbuf);
     break;
   default:
     DIE ("select");
   }
 }

}

void parse_it (int sock, char *rbuf, char **tbuf, char *wbuf) {
  int code, read = 0;

  if (*tbuf - rbuf < 4)		/* assume code */
    return;
      
  if (sscanf (rbuf, "%d%n", &code, &read) == 1) { /* got number */
    if (rbuf + read >= *tbuf) /* must get more */
      return;

    printf ("Return code: %d\n", code);

    if (*(rbuf + read) == '-') {
      if (printIfLine (rbuf, tbuf))
	parse_it (sock, rbuf, tbuf, wbuf);
      return;
    }

    if (code >= 500 && code <= 599) { 
				/* Permanent Negative Completion reply */
      if (code == 500) { /* Syntax error, command unrecognized. */
	/* This may include errors such as command line too long. */
      } else if (code == 501) { /* Syntax error in parameters or arguments. */
      } else if (code == 502) { /* Command not implemented. */
      } else if (code == 503) { /* Bad sequence of commands. */
      } else if (code == 504) { /* Command not implemented for that parameter. */
      } else if (code == 530) { /* Not logged in. */
      } else if (code == 532) { /* Need account for storing files. */
      } else if (code == 550) { /* Requested action not taken. */
	/* File unavailable (e.g., file not found, no access). */
      } else if (code == 551) { /* Requested action aborted: page type unknown. */
      } else if (code == 552) { /* Requested file action aborted. */
	/* Exceeded storage allocation (for current directory or */
	/* dataset). */
      } else if (code == 553) { /* Requested action not taken. */
	/* File name not allowed. */
      } else {
				/* some other reply */
      }
      try_break (sock, wbuf);
    } else if (code >= 400 && code <= 499) { 
				/* Transient Negative Completion reply */
      if (code == 421) { /* Service not available, closing control
			    connection. */
	/* This may be a reply to any command if the service knows it */
	/* must shut down. */
      } else if (code == 425) { /* Can't open data connection. */
      } else if (code == 426) { /* Connection closed; transfer aborted. */
      } else if (code == 450) { /* Requested file action not taken. */
	/* File unavailable (e.g., file busy). */
      } else if (code == 451) { /* Requested action aborted: local error in processing. */
      } else if (code == 452) { /* Requested action not taken. */
	/* Insufficient storage space in system. */
      } else {
      }
      try_break (sock, wbuf);
    } else if (code >= 300 && code <= 399) { 
				/* Positive Intermediate reply */
      if (code == 331) { /* User name okay, need password. */
	if (p_pass != NULL) {
	  char *s;
	  if ((s = malloc (5 + strlen(p_pass) + 2 + 1)) == NULL) {
				/* "PASS " + p_pass + "\r\n"  */
	    DIE ("malloc");
	  }
	  sprintf (s, "PASS %s\r\n", p_pass); /* calculated! */
	  send_it (sock, s, strlen(s));
	  free (s);
	} else 
	  send_it(sock, make_big (wbuf, "PASS ", p_bytes), p_bytes);
      } else if (code == 332) { /* Need account for login. */
      } else if (code == 350) { /* Requested file action pending
				   further information. */
      } else {
      }
    } else if (code >= 200 && code <= 299) { 
				/* Positive Completion reply */
      if (code == 200) { /* Command okay. */
      } else if (code == 202) { /* Command not implemented,
				       superfluous at this site. */
      } else if (code == 211) { /* System status, or system help reply. */
      } else if (code == 212) { /* Directory status. */
      } else if (code == 213) { /* File status. */
      } else if (code == 214) { /* Help message. */
	/* On how to use the server or the meaning of a particular 
	   non-standard command.  This reply is useful only to the 
	   human user. */
      } else if (code == 215) { /* NAME system type. */
	/* Where NAME is an official system name from the list in
	   the  Assigned Numbers document. */
      } else if (code == 220) { /* Service ready for new user. */
	if (p_login != NULL) {
	  char *s;
	  if ((s = malloc (5 + strlen(p_login) + 2 + 1)) == NULL) {
				/* "USER " + p_pass + "\r\n"  */
	    DIE ("malloc");
	  }
	  sprintf (s, "USER %s\r\n", p_login); /* calculated! */
	  send_it (sock, s, strlen(s));
	  free (s);
	} else if (! p_nologin) {
	  send_it(sock, make_big (wbuf, "USER ", p_bytes), p_bytes);
	}
      } else if (code == 221) { /* Service closing control
				   connection. */
	/* Logged out if appropriate. */
	return;
      } else if (code == 225) { /* Data connection open; no
				   transfer in progress. */
      } else if (code == 226) { /* Closing data connection. */
	/* Requested file action successful (for example, file
	   transfer or file abort). */
      } else if (code == 227) { /* Entering Passive Mode (h1,h2,h3,h4,p1,p2). */
      } else if (code == 230) { /* User logged in, proceed. */
	p_state = LOGGED;
	/*	send_it(sock, make_big (wbuf, "HELP ", p_bytes), p_bytes); */
      } else if (code == 250) { /* Requested file action okay, completed. */
      } else if (code == 257) { /* "PATHNAME" created. */
      } else {
      }
      
    } else if (code >= 100 && code <= 199) { 
				/* Positive Preliminary reply */
      if (code == 110) { /* Restart marker reply. */
	
      } else if (code == 120) { /* Service ready in nnn minutes. */
      } else if (code == 125) { /* Data connection already open;
				   transfer starting. */
      } else if (code == 150) { /* File status okay; about to open
				   data connection. */
	
      } else {
      }
    } else {
				/* some other code */
    }
  }
  printIfLine (rbuf, tbuf);
}

char *make_big (char *buf, char *cmd, size_t bytes) {
  int wrote;
  strncpy (buf, cmd, bytes);
  wrote = strlen(buf);
  memset (buf + wrote, 'x', bytes - wrote - 2);
  memcpy (buf + bytes - 2, "\r\n", 2); /* CR LF */
  return buf;
}

int printIfLine (char *rbuf, char **tbuf) {
  char *s;
  int printed = 0;

  if ((s = memchr (rbuf, '\n', *tbuf - rbuf)) != NULL) {
    *s = '\0';
    printf ("Line: %s\n", do_rle(rbuf, strlen(rbuf)));
    memmove (rbuf, s+1, *tbuf - s - 1);
    *tbuf = rbuf + (*tbuf - s - 1);
    printed ++;
  }
  return printed;
}

void usage() {
  fprintf (stderr, "%s", "Usage: bad [options] host\n"
	   "\t[-a] - anonymous login (ftp, email)\n"
	   "\t[-b value] - start scanning at position value\n"
	   "\t[-e value] - stop scnning at position value\n"
	   "\t[-l login] - use login\n"
	   "\t[-n] - no login\n"
	   "\t[-p pass] - use password\n"
	   "\t[-P port] - use port\n"
	   "\t[-s bytes] - max command length\n"
	   "\t[-w sec] - time to wait for respond\n"
	   "\t[-W sec] - time to wait after connect\n"
	   );

  exit (1);
}


int main(int argc, char **argv) {
  char *host = NULL;
  char *buf = NULL;
  char *rbuf = NULL;
  char *tbuf;
  int sock;
  int c;

  while ((c = getopt (argc, argv, "ab:e:s:l:np:P:w:W:")) != EOF) {
    switch (c) {
    case 'a':
      p_login = "ftp";
      p_pass = "bad@whitehats.com";
      break;
    case 'b':
      do {
	int i, v;
	v = atoi(optarg);
	for (i = 0; i != v; i++)
	  if (cmds[i] == NULL)
	    break;
	if (i == v)
	  p_at = v;
      } while (0);
      break;
    case 'e':
      p_end = atoi(optarg);
      break;
    case 's':
      p_bytes = atoi(optarg);
      break;
    case 'l':
      p_login = optarg;
      break;
    case 'n':
      p_nologin = 1;
      break;
    case 'p':
      p_pass = optarg;
      break;
    case 'P':
      p_port = atoi(optarg);
      break;
    case 'w':
      p_wait.tv_sec = atoi (optarg);
      break;
    case 'W':
      p_cwait = atoi (optarg);
      break;
    default:
      fprintf (stderr, "unknown option %s\n", argv[optind-1]);
      usage ();
      break;
    }
  }

  if (argc - optind != 1) {
    usage ();
  }

  host = argv[optind];

  if (p_bytes <= 0 || p_bytes > BUFMAX - 2) {
    fprintf (stderr, "Size (%d) out or range [1..%d]\n",
	     p_bytes, BUFMAX - 2);
    usage ();
  }

  if ((buf = malloc(BUFMAX)) == NULL || (rbuf = malloc(RBUF)) == NULL) {
    DIE("malloc");
  }
  sock = makeconnect (host, p_port);

  if (p_cwait)
    sleep (p_cwait);		/* wait after connect... */

  tbuf = rbuf;
  do_loop (sock, rbuf, &tbuf, buf);

  printIfLine (rbuf, &tbuf);

  close(sock);
  exit(0);
}
